@@ -1,6 +1,7 @@
 #include "Request.h"
 #include <ctime>
 #include <cstdlib>
+#include <type_traits>
 
 namespace panda { namespace protocol { namespace http {
 
@@ -26,35 +27,27 @@ static inline string _method_str (Request::Method rm) {
     }
 }
 
-static inline string generate_boundary(const Request::Form& form) noexcept {
+string Request::_generate_boundary() noexcept {
     const constexpr size_t SZ = (string::MAX_SSO_CHARS / sizeof (int)) + (string::MAX_SSO_CHARS % sizeof (int) == 0 ? 0 : 1);
     const constexpr char alphabet[] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890";
     const constexpr size_t alphabet_sz = sizeof (alphabet) - 1;
     int dices[SZ];
-    bool matches_form;
     string r(40, '-');
-    do {
-        for(size_t i = 0; i <SZ; ++i) { dices[i] = std::rand(); }
+    for(size_t i = 0; i <SZ; ++i) { dices[i] = std::rand(); }
 
-        const char* random_bytes = (const char*)dices;
-        for(size_t i = r.size() - 17; i < r.size(); ++i) {
-            r[i] = alphabet[*random_bytes++ % alphabet_sz];
-        }
+    const char* random_bytes = (const char*)dices;
+    for(size_t i = r.size() - 17; i < r.size(); ++i) {
+        r[i] = alphabet[*random_bytes++ % alphabet_sz];
+    }
 
-        matches_form = false;
-        for(auto it = form.begin(); (!matches_form) && (it != form.end()); ++it) {
-            matches_form = (it->first.find(r) != string::npos) || (it->second.find(r) != string::npos);
-        }
-    } while(matches_form);
     return r;
 }
 
 Request::Method Request::method() const noexcept {
     if (_method == Method::unspecified) {
-        if (form && form.enc_type() == EncType::MULTIPART && (!form.empty() || (uri && !uri->query().empty()))) {
-            return Method::POST;
-        }
-        return Method::GET;
+        bool use_post = (form && form.enc_type() == EncType::MULTIPART && (!form.empty() || (uri && !uri->query().empty()))) // complete form
+                     || _form_streaming != FormStreaming::none;
+        return use_post ? Method::POST : Method::GET;
     }
     return _method;
 }
@@ -75,7 +68,7 @@ string Request::_http_header (SerializationContext& ctx) const {
     auto tmp_http_ver = !ctx.http_version ? 11 : ctx.http_version;
     string out_content_length;
     bool calc_content_length
-              = !chunked
+              = !ctx.chunked
             && (ctx.body->parts.size() || body_method)
             && !headers.has("Content-Length");
     if (calc_content_length) out_content_length = panda::to_string(ctx.body->length());
@@ -193,22 +186,33 @@ string Request::_http_header (SerializationContext& ctx) const {
 
 std::vector<string> Request::to_vector () const {
     SerializationContext ctx;
-    ctx.compression = compression.type;
+    if (_form_streaming != FormStreaming::none && _form_streaming != FormStreaming::started)
+        throw "form streaming wasn't finished";
+
+    bool form_streaming =  _form_streaming == FormStreaming::started;
+    /* it seems nobody supports muliptart + gzip + chunk */
+    ctx.compression = !form_streaming ? compression.type : Compression::Type::IDENTITY;
     ctx.body        = &body;
     ctx.uri         = uri.get();
+    ctx.chunked     = this->chunked || form_streaming;
+
+    auto add_form_header = [&](auto& boundary) {
+        string ct = "multipart/form-data; boundary=";
+        ct += boundary;
+        ctx.handled_headers.add("Content-Type", ct);
+    };
 
     Body form_body;
     URI form_uri;
     if (form) {
         if (form.enc_type() == EncType::MULTIPART) {
             if (!form.empty() || (uri && !uri->query().empty())) {
-                auto boundary = generate_boundary(form);
-                form.to_body(form_body, form_uri, uri, boundary);
-                string ct = "multipart/form-data; boundary=";
-                ct += boundary;
-                ctx.handled_headers.add("Content-Type", ct);
-                ctx.body     = &form_body;
-                ctx.uri      = &form_uri;
+                auto boundary = form_streaming ? _form_boundary : _generate_boundary();
+                ctx.uri  = form.to_body(form_body, form_uri, uri, boundary);
+                ctx.body = &form_body;
+                add_form_header(boundary);
+            } else if (form_streaming) {
+                add_form_header(_form_boundary);
             }
         }
         else if((form.enc_type() == EncType::URLENCODED) && !form.empty()) {
@@ -216,6 +220,9 @@ std::vector<string> Request::to_vector () const {
             ctx.uri = &form_uri;
         }
     }
+    else if (form_streaming) {
+        add_form_header(_form_boundary);
+    }
     return _to_vector(ctx, [&]() { return _compile_prepare(ctx); }, [&]() { return _http_header(ctx); });
 }
 
@@ -234,51 +241,213 @@ std::uint8_t Request::allowed_compression (bool inverse) const noexcept {
     return result;
 }
 
-void Request::Form::to_body(Body& body, uri::URI &uri, const uri::URISP original_uri, const string &boundary) const noexcept {
-    using Container = string_multimap<string, string>;
+static string form_trailer(const string& boundary) noexcept {
+    auto sz = boundary.size() + 6;
+    string r(sz);
+    r += "--";
+    r += boundary;
+    r += "--\r\n";
+    return r;
+}
 
-    auto serialize = [&body, &boundary](auto fields_count, const Container& container) {
-        // pass 1: calc total
-        size_t size = (
-                boundary.length() + 2   /* \r\n */
-                + 37 + 2                /* Content-Disposition: form-data; name="" + \r\n */
-            ) * fields_count  + 2;      /* -- */
-        for(auto it : container) size += it.second.length();
+namespace tag {
+    using uri = std::integral_constant<int, 0>;
+    using form = std::integral_constant<int, 1>;
+}
 
-        // pass 2: merge strings
-        string r(size);
-        for(auto it : container) {
-            r += boundary;
-            r += "\r\n";
-            r += "Content-Disposition: form-data; name=\"";
-            r += it.first;
-            r += "\"\r\n";
-            r += "\r\n";
-            r += it.second;
-            r += "\r\n";
-        }
-        r += boundary;
-        r += "--\r\n";
+template<typename Tag> struct Helper;
 
-        body.parts.emplace_back(r);
+template<> struct Helper<tag::form> {
+    using Field =  Request::Form::value_type;
+
+    struct PatrialField {
+        string name;
+        string mime_type;
+        string filename;
+        bool complete = false;
     };
 
-    if (empty()) {
-        auto& q = original_uri->query();
-        serialize(q.size(), q);
-        uri = URI(*original_uri);
-        uri.query().clear();
-    } else {
-        serialize(size(), *this);
+    struct FullField: PatrialField {
+        FullField(const string& name_, const string& filename_, const string& mime_type_, const string& value_):
+            PatrialField{name_, mime_type_, filename_, true}, value{value_}{}
+        string value;
+    };
+
+    static size_t buffer_size(const string &boundary, const Request::Form& container) noexcept {
+        auto fields_count = container.size();
+        size_t size = (
+                boundary.length() + 4   /* "--" prefix and "\r\n" */
+                + 37 + 2                /* Content-Disposition: form-data; name="" + \r\n */
+            ) * fields_count  + 2;      /* -- */
+        for(auto it : container) {
+            size += it.second.value.length();
+            auto& name = it.second.name;
+            if (name) {
+                size += name.length() + 14;   //; filename=""
+            }
+            auto& ct = it.second.content_type;
+            if (ct) {
+                size += ct.size() + 18; //Content-Type: xxx\r\n
+            }
+        }
+        return size;
+    }
+
+    static void append_header(string& r, const string& header, const string& value) noexcept {
+        r += header;
+        r += ": ";
+        r += value;
+        r += "\r\n";
+    }
+
+    static void append(string& r, const PatrialField& field, const string& boundary) noexcept {
+        r += "--";
+        r += boundary;
+        r += "\r\n";
+        r += "Content-Disposition: form-data; name=\"";
+        r += field.name;
+        r += "\"";
+        if (field.filename) {
+            r+= "; filename=\"";
+            r += field.filename;
+            r += "\"";
+        }
+        r += "\r\n";
+
+        if (field.mime_type) {
+            r += "Content-Type: ";
+            r += field.mime_type;
+            r += "\r\n";
+        }
+
+        if (field.complete) r += "\r\n";
+    }
+
+    static void append(string& r, const FullField& field, const string& boundary) noexcept {
+        append(r, (const PatrialField&)field, boundary);
+        r += field.value;
+        r += "\r\n";
+    }
+
+    static void append(string& r, const Field& it, const string& boundary) noexcept {
+        append(r, FullField{it.first, it.second.name, it.second.content_type, it.second.value}, boundary);
+    }
+
+};
+
+template<> struct Helper<tag::uri> {
+    using Field =  string_multimap<string, string>::value_type;
+
+    static size_t buffer_size(const string &boundary, const string_multimap<string, string>& container) noexcept {
+        auto fields_count= container.size();
+        size_t size = (
+                boundary.length() + 4   /* "--" prefix and "\r\n" */
+                + 37 + 2                /* Content-Disposition: form-data; name="" + \r\n */
+            ) * fields_count  + 2;      /* -- */
+        for(auto it : container) {
+            size += it.second.length();
+        }
+        return size;
+    }
+
+    static void append(string& r, const Field& it, const string& boundary) noexcept {
+        r += "--";
+        r += boundary;
+        r += "\r\n";
+        r += "Content-Disposition: form-data; name=\"";
+        r += it.first;
+        r += "\"";
+        r += "\r\n";
+        r += "\r\n";
+        r += it.second;
+        r += "\r\n";
+    }
+};
+
+void Request::form_file_finalize(string& out) noexcept {
+    if (_form_streaming == FormStreaming::file) {
+        if (compressor) {
+            out += compressor->flush();
+            compressor.reset();
+        }
+        out += "\r\n"; /* finalize file */
     }
 }
 
-void Request::Form::to_uri  (uri::URI &uri, const URISP original_uri) const noexcept {
+
+Request::wrapped_chunk Request::form_finish() {
+    if (_form_streaming == FormStreaming::none) throw "form streaming was not started";
+    if (_form_streaming == FormStreaming::done) throw "form streaming already complete";
+
+    string data;
+    form_file_finalize(data);
+    data += form_trailer(_form_boundary);
+    _form_streaming = FormStreaming::done;
+    return final_chunk(data);
+}
+
+Request::wrapped_chunk Request::form_field(const string& name, const string& content, const string& filename, const string& mime_type) {
+    using H = Helper<tag::form>;
+    if (_form_streaming == FormStreaming::none) throw "form streaming was not started";
+    if (_form_streaming == FormStreaming::done) throw "form streaming already complete";
+
+    string data;
+    form_file_finalize(data);
+    H::append(data, H::FullField{name, filename, mime_type, content}, _form_boundary);
+    return make_chunk(data, compression::CompressorPtr{});  // we don't compress
+}
+
+Request::wrapped_chunk Request::form_file(const string& name, const string filename, const string& mime_type) {
+    using H = Helper<tag::form>;
+    if (_form_streaming == FormStreaming::none) throw "form streaming was not started";
+    if (_form_streaming == FormStreaming::done) throw "form streaming already complete";
+
+    string data;
+    form_file_finalize(data);
+    _form_streaming = FormStreaming::file;
+
+    H::append(data, H::PatrialField{name, mime_type, filename, false}, _form_boundary);
+    data += "\r\n";
+
+    return make_chunk(data, compression::CompressorPtr{});  // we don't compress
+}
+
+Request::wrapped_chunk Request::form_data(const string& content) {
+    if (_form_streaming != FormStreaming::file) throw "form file streaming was not started";
+    return make_chunk(content);
+}
+
+template<typename Tag, typename Container>
+void _serialize(Body& body, const string &boundary, const Container& container) {
+    using H = Helper<Tag>;
+    string r(H::buffer_size(boundary, container));
+    for(auto it : container) { H::append(r, it, boundary); }
+    r += form_trailer(boundary);
+    body.parts.emplace_back(r);
+}
+
+const uri::URI *Request::Form::to_body(Body& body, uri::URI &uri, const uri::URISP original_uri, const string &boundary) const noexcept {
+    if (empty()) {
+        if (!original_uri) return {};
+        auto& q = original_uri->query();
+        _serialize<tag::uri>(body, boundary, q);
+        uri = URI(*original_uri);
+        uri.query().clear();
+        return &uri;
+    } else {
+        _serialize<tag::form>(body, boundary, *this);
+        return original_uri.get();
+    }
+}
+
+void Request::Form::to_uri  (uri::URI &uri, const URISP original_uri) const  {
     if (original_uri) uri = *original_uri;
     else              uri = "/";
     auto& q = uri.query();
     for(auto it : *this) {
-        q.insert({it.first, it.second});
+        auto& named = it.second;
+        if (named.name) throw string("form contains named field (filename) " + it.second.value + ", it cannot be converted to URI");
+        q.insert({it.first, it.second.value});
     }
 }
 
